/*
 * Copyright © 2005 Novell, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Novell, Inc. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Novell, Inc. makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * NOVELL, INC. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL NOVELL, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: David Reveman <davidr@novell.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>

#include <beryl.h>

#define HOME_PLUGINDIR ".beryl/plugins"

CompPlugin *plugins = 0;

static Bool
dlloaderLoadPlugin (CompPlugin * p, char *path, char *name)
{
    char *file;
    void *dlhand;

    file = malloc ((path ? strlen (path) : 0) + strlen (name) + 8);
    if (!file)
        return FALSE;

    if (path)
        sprintf (file, "%s/lib%s.so", path, name);
    else
        sprintf (file, "lib%s.so", name);

    dlhand = dlopen (file, RTLD_LAZY);
    if (dlhand)
    {
        PluginGetInfoProc getInfo;
        char *error;

        dlerror ();

        getInfo = (PluginGetInfoProc) dlsym (dlhand, "getCompPluginInfo");

        error = dlerror ();
        if (error)
        {
            fprintf (stderr, "%s: dlsym: %s\n", programName, error);

            getInfo = 0;
        }

        if (getInfo)
        {
            p->vTable = (*getInfo) ();
            if (!p->vTable)
            {
                fprintf (stderr, _("%s: Couldn't get vtable from '%s' plugin\n"),
                         programName, file);

                dlclose (dlhand);
                free (file);

                return FALSE;
            }
            if (p->vTable->version != BERYL_VERSION ||
                p->vTable->struct_plugin_size != sizeof (CompPlugin) ||
                p->vTable->struct_display_size != sizeof (CompDisplay) ||
                p->vTable->struct_screen_size != sizeof (CompScreen) ||
                p->vTable->struct_window_size != sizeof (CompWindow) ||
                p->vTable->struct_texture_size != sizeof (CompTexture) ||
                p->vTable->struct_icon_size != sizeof (CompIcon))
            {

                fprintf (stderr,
                         _("%s: '%s' plugin version does not match beryl version.\n"),
                         programName, file);

                dlclose (dlhand);
                free (file);

                return FALSE;
            }
        }
        else
        {
            fprintf (stderr, _("%s: Failed to lookup getCompPluginInfo in '%s' ")
                     "plugin\n", programName, file);

            dlclose (dlhand);
            free (file);

            return FALSE;
        }
    }
    else
    {
        free (file);

        return FALSE;
    }

    p->devPrivate.ptr = dlhand;
    p->devType = "dlloader";

    return TRUE;
}

static void
dlloaderUnloadPlugin (CompPlugin * p)
{
    dlclose (p->devPrivate.ptr);
}

LoadPluginProc loaderLoadPlugin = dlloaderLoadPlugin;
UnloadPluginProc loaderUnloadPlugin = dlloaderUnloadPlugin;

Bool
initPluginForDisplay (CompPlugin * p, CompDisplay * d)
{
    CompScreen *s, *failedScreen = d->screens;
    Bool status = TRUE;

    if (!(*p->vTable->initDisplay) (p, d))
        return FALSE;

    for (s = d->screens; s; s = s->next)
    {
        if (!(*s->initPluginForScreen) (p, s))
        {
            fprintf (stderr, _("%s: Plugin '%s':initScreen failed\n"),
                     programName, p->vTable->name);
            failedScreen = s;
            status = FALSE;
            break;
        }
    }

    for (s = d->screens; s != failedScreen; s = s->next)
        (*s->finiPluginForScreen) (p, s);

    return status;
}

void
finiPluginForDisplay (CompPlugin * p, CompDisplay * d)
{
    CompScreen *s;

    for (s = d->screens; s; s = s->next)
        (*s->finiPluginForScreen) (p, s);

    (*p->vTable->finiDisplay) (p, d);
}

Bool
initPluginForScreen (CompPlugin * p, CompScreen * s)
{
    Bool status = TRUE;

    if (p->vTable->initScreen)
    {
        if (!(*p->vTable->initScreen) (p, s))
            return FALSE;
    }

    if (p->vTable->initWindow)
    {
        CompWindow *w, *failedWindow = s->windows;

        for (w = s->windows; w; w = w->next)
        {
            if (!(*p->vTable->initWindow) (p, w))
            {
                fprintf (stderr, _("%s: Plugin '%s':initWindow "
                         "failed\n"), programName, p->vTable->name);
                failedWindow = w;
                status = FALSE;
                break;
            }
        }

        if (p->vTable->finiWindow)
        {
            for (w = s->windows; w != failedWindow; w = w->next)
                (*p->vTable->finiWindow) (p, w);
        }
    }

    return status;
}

void
finiPluginForScreen (CompPlugin * p, CompScreen * s)
{
    if (p->vTable->finiWindow)
    {
        CompWindow *w = s->windows;

        for (w = s->windows; w; w = w->next)
            (*p->vTable->finiWindow) (p, w);
    }

    if (p->vTable->finiScreen)
        (*p->vTable->finiScreen) (p, s);
}

static Bool
initPlugin (CompPlugin * p)
{
    CompDisplay *d = compDisplays;

    if (!(*p->vTable->init) (p))
    {
        fprintf (stderr, _("%s: InitPlugin '%s' failed\n"), programName,
                 p->vTable->name);
        return FALSE;
    }

    if (d)
    {
        if (!(*d->initPluginForDisplay) (p, d))
        {
            fprintf (stderr, _("%s: Plugin '%s':initDisplay failed\n"),
                     programName, p->vTable->name);

            (*p->vTable->fini) (p);

            return FALSE;
        }
    }

    return TRUE;
}

static void
finiPlugin (CompPlugin * p)
{
    CompDisplay *d = compDisplays;

    if (d)
        (*d->finiPluginForDisplay) (p, d);

    (*p->vTable->fini) (p);
}

void
screenInitPlugins (CompScreen * s)
{
    CompPlugin *p;
    int i, j = 0;

    for (p = plugins; p; p = p->next)
        j++;

    while (j--)
    {
        i = 0;
        for (p = plugins; i < j; p = p->next)
            i++;

        if (p->vTable->initScreen)
            (*s->initPluginForScreen) (p, s);
    }
}

void
screenFiniPlugins (CompScreen * s)
{
    CompPlugin *p;

    for (p = plugins; p; p = p->next)
    {
        if (p->vTable->finiScreen)
            (*s->finiPluginForScreen) (p, s);
    }
}

void
windowInitPlugins (CompWindow * w)
{
    CompPlugin *p;

    for (p = plugins; p; p = p->next)
    {
        if (p->vTable->initWindow)
            (*p->vTable->initWindow) (p, w);
    }
}

void
windowFiniPlugins (CompWindow * w)
{
    CompPlugin *p;

    for (p = plugins; p; p = p->next)
    {
        if (p->vTable->finiWindow)
            (*p->vTable->finiWindow) (p, w);
    }
}

CompPlugin *
findActivePlugin (char *name)
{
    CompPlugin *p;

    for (p = plugins; p; p = p->next)
    {
        if (strcmp (p->vTable->name, name) == 0)
            return p;
    }

    return 0;
}

void
unloadPlugin (CompPlugin * p)
{
    (*loaderUnloadPlugin) (p);
    free (p);
}

CompPlugin *
loadPlugin (char *name)
{
    CompPlugin *p;
    char *home, *plugindir;
    Bool status;

    p = malloc (sizeof (CompPlugin));
    if (!p)
        return 0;

    p->next = 0;
    p->devPrivate.uval = 0;
    p->devType = NULL;
    p->vTable = 0;

    home = getenv ("HOME");
    if (home)
    {
        plugindir = malloc (strlen (home) + strlen (HOME_PLUGINDIR) + 3);
        if (plugindir)
        {
            sprintf (plugindir, "%s/%s", home, HOME_PLUGINDIR);
            status = (*loaderLoadPlugin) (p, plugindir, name);
            free (plugindir);

            if (status)
                return p;
        }
    }

    status = (*loaderLoadPlugin) (p, PLUGINDIR, name);
    if (status)
        return p;

    status = (*loaderLoadPlugin) (p, NULL, name);
    if (status)
        return p;

    fprintf (stderr, _("%s: Couldn't load plugin '%s'\n"), programName, name);

    return 0;
}

static Bool
checkPluginDeps (CompPlugin * p)
{
    CompPluginDep *deps;
    int        nDeps;
    Bool hasAfterAltDeps = FALSE;
    Bool afterAltDepsOk = TRUE;

    // Conflict check
    CompPlugin *plugin;
    for ( plugin = plugins; plugin; plugin = plugin->next )
    {
        deps = plugin->vTable->deps;
        nDeps = plugin->vTable->nDeps;
        while ( nDeps -- )
        {
            if ( deps->rule == CompPluginRuleConflicts )
            {
                if ( strcmp (p->vTable->name, deps->plugin) == 0 )
                {
                    fprintf ( stderr, _("%s: '%s' plugin conflicts with '%s' "
                             "plugin\n"), programName, p->vTable->name, plugin->vTable->name );
                    return FALSE;
                }
            }
            deps++;
        }
    }

    deps = p->vTable->deps;
    nDeps = p->vTable->nDeps;

    while (nDeps--)
    {
        switch (deps->rule) {
        case CompPluginRuleAfterAlt:
            if ( ! hasAfterAltDeps )
            {
                    hasAfterAltDeps = TRUE;
                    afterAltDepsOk = FALSE;
            }
            else if ( afterAltDepsOk )
                    break;
            if ( findActivePlugin ( deps->plugin ) )
                afterAltDepsOk = TRUE;
            break;
        case CompPluginRuleConflicts:
            if (findActivePlugin (deps->plugin))
            {
                fprintf (stderr, _("%s: '%s' plugin conflicts with '%s' "
                         "plugin\n"), programName, p->vTable->name, deps->plugin);

                return FALSE;
            }
            break;
        case CompPluginRuleBefore:
            if (findActivePlugin (deps->plugin))
            {
                fprintf (stderr, _("%s: '%s' plugin must be loaded before '%s' "
                         "plugin\n"), programName, p->vTable->name,
                         deps->plugin);

                return FALSE;
            }
            break;
        case CompPluginRuleAfter:
            if (!findActivePlugin (deps->plugin))
            {
                fprintf (stderr, _("%s: '%s' plugin must be loaded after '%s' "
                         "plugin\n"), programName, p->vTable->name,
                         deps->plugin);

                return FALSE;
            }
            break;
        }

        deps++;
    }

    if ( afterAltDepsOk == FALSE )
    {
        fprintf (stderr, _("%s: '%s' plugin was not loaded due to unfetched "
                 "dependencies\n"), programName, p->vTable->name);
        return FALSE;
    }

    return TRUE;
}

static Bool splashWasActive = FALSE;

static void splashActivate(CompPlugin *p)
{
    CompDisplay *d = compDisplays;
    if (splashWasActive || strcmp(p->vTable->name,"splash") != 0)
        return;
    splashWasActive = TRUE;

    int i,nOpts = 0;
    CompOption *o = p->vTable->getDisplayOptions(d,&nOpts);
    CompScreen *s;

    for (s = d->screens; s; s = s->next) {
        for (i = 0; i < nOpts; i++) {
            if (strcmp(o[i].name,"initiate") == 0)
            {
                CompOption root;
                root.type = CompOptionTypeInt;
                root.name = "root";
                root.value.i = s->root;
                printf("Initiating splash\n");
                (* o[i].value.action.initiate) (d, &(o[i].value.action), 0, &root, 1);
            }
        }
    }
}

Bool
pushPlugin (CompPlugin * p)
{
    if (findActivePlugin (p->vTable->name))
    {
        fprintf (stderr, _("%s: Plugin '%s' already active\n"), programName,
                 p->vTable->name);

        return FALSE;
    }

    if (!checkPluginDeps (p))
    {
        fprintf (stderr, _("%s: Can't activate '%s' plugin due to dependency "
                 "problems\n"), programName, p->vTable->name);

        return FALSE;
    }

    p->next = plugins;
    plugins = p;

    if (!initPlugin (p))
    {
        fprintf (stderr, _("%s: Couldn't activate plugin '%s'\n"), programName,
                 p->vTable->name);
        plugins = p->next;

        return FALSE;
    }

    splashActivate(p);

    return TRUE;
}

CompPlugin *
popPlugin (void)
{
    CompPlugin *p = plugins;

    if (!p)
        return 0;

    finiPlugin (p);

    plugins = p->next;

    return p;
}

CompPlugin *
getPlugins (void)
{
    return plugins;
}
